""" Starting beta stable version 0.9

 - new in version 0.9.1
   - read MCNP macrobodies : RPP, BOX, RCC, REC, TRC
   - read SQ surfaces
   - hyperboloid 1 or 2 sheets (revolution), paraboloid (revolution), elliptical and hyperbolic cylinder, 
     elliptical cone and elliptical torus surfaces have been added.

   Changes have not been fully checked, some bugs may still remains. Please contact the authors in case of failure.
"""

GEOReverse_Version = "0.9.1"
GEOReverse_ReleaseDate = "16/11/2023"
